'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

//      
var toPath = function toPath(key) {
  if (key === null || key === undefined) {
    return [];
  }
  if (typeof key !== 'string') {
    throw new Error('toPath() expects a string');
  }
  return key.length ? key.split(/[.[\]]+/).filter(Boolean) : [];
};

//      
var getIn = function getIn(state, complexKey) {
  // Intentionally using iteration rather than recursion
  var path = toPath(complexKey);
  var current = state;
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = path[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var key = _step.value;

      if (current === undefined || current === null || !isNaN(current)) {
        return undefined;
      }
      if (Array.isArray(current) && isNaN(key)) {
        return undefined; // Bad user. 404 Not Found. Can't read a non-numeric key from an array.
      }
      current = current[key];
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return current;
};

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};



















var defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};

var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};



































var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

//      
var setInRecursor = function setInRecursor(current, index, path, value) {
  if (index >= path.length) {
    // end of recursion
    return value;
  }
  var key = path[index];

  // determine type of key
  if (isNaN(key)) {
    // object set
    if (current === undefined || current === null) {
      // recurse
      var _result2 = setInRecursor(undefined, index + 1, path, value);

      // delete or create an object
      return _result2 === undefined ? undefined : defineProperty({}, key, _result2);
    }
    if (Array.isArray(current)) {
      throw new Error('Cannot set a non-numeric property on an array');
    }
    // current exists, so make a copy of all its values, and add/update the new one
    var _result = setInRecursor(current[key], index + 1, path, value);
    var numKeys = Object.keys(current).length;
    return _result === undefined && (current[key] === undefined && numKeys === 0 || // object was already empty
    current[key] !== undefined && numKeys <= 1) // only key we had was the one we are deleting
    ? undefined : _extends({}, current, defineProperty({}, key, _result));
  }
  // array set
  var numericKey = Number(key);
  if (current === undefined || current === null) {
    // recurse
    var _result3 = setInRecursor(undefined, index + 1, path, value);

    // if nothing returned, delete it
    if (_result3 === undefined) {
      return undefined;
    }

    // create an array
    var _array = [];
    _array[numericKey] = _result3;
    return _array;
  }
  if (!Array.isArray(current)) {
    throw new Error('Cannot set a numeric property on an object');
  }
  // recurse
  var existingValue = current[numericKey];
  var result = setInRecursor(existingValue, index + 1, path, value);

  // if nothing returned and nothing in current, delete it
  if (result === undefined && (existingValue === undefined && current.length === 0 || existingValue !== undefined && current.length === 1)) {
    return undefined;
  }

  // current exists, so make a copy of all its values, and add/update the new one
  var array = [].concat(toConsumableArray(current));
  if (result === undefined) {
    array.splice(numericKey, 1);
  } else {
    array[numericKey] = result;
  }
  return array;
};

var setIn = function setIn(state, key, value) {
  if (state === undefined || state === null) {
    throw new Error('Cannot call setIn() with ' + String(state) + ' state');
  }
  if (key === undefined || key === null) {
    throw new Error('Cannot call setIn() with ' + String(key) + ' key');
  }
  // Recursive function needs to accept and return State, but public API should
  // only deal with Objects
  return setInRecursor(state, 0, toPath(key), value);
};

//      


/**
 * Converts internal field state to published field state
 */
var publishFieldState = function publishFieldState(form, field) {
  var submitFailed = form.submitFailed,
      submitSucceeded = form.submitSucceeded;
  var active = field.active,
      blur = field.blur,
      change = field.change,
      error = field.error,
      focus = field.focus,
      initial = field.initial,
      name = field.name,
      submitError = field.submitError,
      touched = field.touched,
      value = field.value,
      visited = field.visited;

  var pristine = initial === value;
  var valid = !error && !submitError;
  return {
    active: active,
    blur: blur,
    change: change,
    dirty: !pristine,
    error: error,
    focus: focus,
    initial: initial,
    invalid: !valid,
    name: name,
    pristine: pristine,
    submitError: submitError,
    submitFailed: submitFailed,
    submitSucceeded: submitSucceeded,
    touched: touched,
    valid: valid,
    value: value,
    visited: visited
  };
};

//      
var fieldSubscriptionItems = ['active', 'dirty', 'error', 'initial', 'invalid', 'pristine', 'submitError', 'submitFailed', 'submitSucceeded', 'touched', 'valid', 'value', 'visited'];

//      
var subscriptionFilter = function (dest, src, previous, subscription, keys) {
  var different = false;
  keys.forEach(function (key) {
    if (subscription[key]) {
      dest[key] = src[key];
      if (!previous || src[key] !== previous[key]) {
        different = true;
      }
    }
  });
  return different;
};

//      
/**
 * Filters items in a FieldState based on a FieldSubscription
 */
var filterFieldState = function filterFieldState(state, previousState, subscription, force) {
  var result = {
    blur: state.blur,
    change: state.change,
    focus: state.focus,
    name: state.name
  };
  var different = subscriptionFilter(result, state, previousState, subscription, fieldSubscriptionItems) || !previousState;
  return different || force ? result : undefined;
};

//      
var formSubscriptionItems = ['active', 'dirty', 'error', 'initialValues', 'invalid', 'pristine', 'submitting', 'submitError', 'submitFailed', 'submitSucceeded', 'valid', 'validating', 'values'];

//      
/**
 * Filters items in a FormState based on a FormSubscription
 */
var filterFormState = function filterFormState(state, previousState, subscription, force) {
  var result = {};
  var different = subscriptionFilter(result, state, previousState, subscription, formSubscriptionItems) || !previousState;
  return different || force ? result : undefined;
};

//      
var shallowEqual = function shallowEqual(a, b) {
  if (a === b) {
    return true;
  }
  if ((typeof a === 'undefined' ? 'undefined' : _typeof(a)) !== 'object' || !a || (typeof b === 'undefined' ? 'undefined' : _typeof(b)) !== 'object' || !b) {
    return false;
  }
  var keysA = Object.keys(a);
  var keysB = Object.keys(b);
  if (keysA.length !== keysB.length) {
    return false;
  }
  var bHasOwnProperty = Object.prototype.hasOwnProperty.bind(b);
  for (var idx = 0; idx < keysA.length; idx++) {
    var key = keysA[idx];
    if (!bHasOwnProperty(key) || a[key] !== b[key]) {
      return false;
    }
  }
  return true;
};

//      
var memoize = function memoize(fn) {
  var lastArgs = void 0;
  var lastResult = void 0;
  return function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    if (!lastArgs || args.length !== lastArgs.length || args.some(function (arg, index) {
      return !shallowEqual(lastArgs[index], arg);
    })) {
      lastArgs = args;
      lastResult = fn.apply(undefined, args);
    }
    return lastResult;
  };
};

var isPromise = (function (obj) {
  return !!obj && ((typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' || typeof obj === 'function') && typeof obj.then === 'function';
});

//      
var FORM_ERROR = Symbol('form-error');
var version = '0.0.2';

var safeFormStateCast = function safeFormStateCast(_ref) {
  var active = _ref.active,
      dirty = _ref.dirty,
      error = _ref.error,
      invalid = _ref.invalid,
      initialValues = _ref.initialValues,
      pristine = _ref.pristine,
      submitting = _ref.submitting,
      submitFailed = _ref.submitFailed,
      submitSucceeded = _ref.submitSucceeded,
      submitError = _ref.submitError,
      valid = _ref.valid,
      validating = _ref.validating,
      values = _ref.values;
  return {
    active: active,
    dirty: dirty,
    error: error,
    invalid: invalid,
    initialValues: initialValues,
    pristine: pristine,
    submitting: submitting,
    submitFailed: submitFailed,
    submitSucceeded: submitSucceeded,
    submitError: submitError,
    valid: valid,
    validating: validating,
    values: values
  };
};

var safeFieldStateCast = function safeFieldStateCast(_ref2) {
  var active = _ref2.active,
      blur = _ref2.blur,
      change = _ref2.change,
      error = _ref2.error,
      focus = _ref2.focus,
      initial = _ref2.initial,
      lastFieldState = _ref2.lastFieldState,
      name = _ref2.name,
      submitError = _ref2.submitError,
      touched = _ref2.touched,
      value = _ref2.value,
      visited = _ref2.visited;

  var pristine = value === initial;
  var invalid = error || submitError;
  return {
    active: active,
    blur: blur,
    change: change,
    dirty: !pristine,
    error: error,
    focus: focus,
    initial: initial,
    invalid: invalid,
    name: name,
    pristine: pristine,
    submitError: submitError,
    touched: touched,
    valid: !invalid,
    value: value,
    visited: visited
  };
};

function notifySubscriber(subscriber, subscription, state, lastState, filter) {
  var force = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : false;

  var notification = filter(state, lastState, subscription, force);
  if (notification) {
    subscriber(notification);
  }
}

function notify(_ref3, state, lastState, filter) {
  var entries = _ref3.entries;

  Object.keys(entries).forEach(function (key) {
    var _entries$Number = entries[Number(key)],
        subscription = _entries$Number.subscription,
        subscriber = _entries$Number.subscriber;

    notifySubscriber(subscriber, subscription, state, lastState, filter);
  });
}

var createForm = function createForm(config) {
  if (!config) {
    throw new Error('No config specified');
  }
  var debug = config.debug,
      initialValues = config.initialValues,
      onSubmit = config.onSubmit,
      validate = config.validate,
      validateOnBlur = config.validateOnBlur;

  if (!onSubmit) {
    throw new Error('No onSubmit function specified');
  }
  var formState = {
    dirty: false,
    initialValues: initialValues && _extends({}, initialValues),
    invalid: false,
    pristine: true,
    submitting: false,
    submitFailed: false,
    submitSucceeded: false,
    valid: true,
    validating: false,
    values: initialValues ? _extends({}, initialValues) : {}
  };
  var state = {
    subscribers: { index: 0, entries: {} },
    fieldSubscribers: {},
    fields: {},
    formState: formState,
    lastFormState: undefined,
    validating: 0
  };
  var inBatch = false;

  var runRecordLevelValidation = function runRecordLevelValidation(setError) {
    var promises = [];
    if (validate) {
      var processErrors = function processErrors(errors) {
        state.error = errors[FORM_ERROR];
        Object.keys(state.fields).forEach(function (key) {
          setError(key, getIn(errors, key));
        });
      };
      var errorsOrPromise = validate(state.formState.values);
      if (isPromise(errorsOrPromise)) {
        promises.push(errorsOrPromise.then(processErrors));
      } else {
        processErrors(errorsOrPromise);
      }
    }
    return promises;
  };

  var runFieldLevelValidation = function runFieldLevelValidation(field, setError) {
    var validators = field.validators;

    var promises = [];
    var validatorKeys = Object.keys(validators);
    if (validatorKeys.length) {
      var error = void 0;
      Object.keys(validators).forEach(function (index) {
        var validator = validators[Number(index)];
        var errorOrPromise = validator(field.value, state.formState.values);
        if (errorOrPromise && isPromise(errorOrPromise)) {
          promises.push(errorOrPromise.then(setError));
        } else if (!error) {
          // first registered validator wins
          error = errorOrPromise;
        }
      });
      setError(error);
    }
    return promises;
  };

  var runValidation = function runValidation(callback) {
    var fields = state.fields;

    var fieldKeys = Object.keys(fields);
    if (!validate && !fieldKeys.some(function (key) {
      return fields[key].validators && Object.keys(fields[key].validators).length;
    })) {
      if (callback) {
        callback();
      }
      return; // no validation rules
    }

    // sync version of setError
    var recordLevelErrors = {};
    var fieldLevelErrors = {};
    var setRecordLevelError = function setRecordLevelError(name, error) {
      recordLevelErrors[name] = error;
    };
    var setFieldLevelError = function setFieldLevelError(name, error) {
      fieldLevelErrors[name] = error;
    };

    var promises = [].concat(toConsumableArray(runRecordLevelValidation(function (name, error) {
      return setRecordLevelError(name, error);
    })), toConsumableArray(fieldKeys.reduce(function (result, name) {
      return [].concat(toConsumableArray(result), toConsumableArray(runFieldLevelValidation(fields[name], function (error) {
        return setFieldLevelError(name, error);
      })));
    }, [])));

    // process sync errors
    fieldKeys.forEach(function (name) {
      // field-level errors take precedent over record-level errors
      fields[name].error = fieldLevelErrors[name] || recordLevelErrors[name];
    });

    if (promises.length) {
      // sync errors have been set. notify listeners while we wait for others
      state.validating++;
      if (callback) {
        callback();
      }

      // reassign setError functions for async responses
      setRecordLevelError = setFieldLevelError = function setFieldLevelError(name, error) {
        fields[name].error = error;
      };

      Promise.all(promises).then(function () {
        state.validating--;
        if (callback) {
          callback();
        }
      });
    } else if (callback) {
      callback();
    }
  };

  var notifyFieldListeners = function notifyFieldListeners(force) {
    if (inBatch) {
      return;
    }
    var fields = state.fields,
        fieldSubscribers = state.fieldSubscribers,
        formState = state.formState;

    Object.keys(fields).forEach(function (name) {
      var field = fields[name];
      var fieldState = publishFieldState(formState, field);
      var lastFieldState = field.lastFieldState;

      if (!shallowEqual(fieldState, lastFieldState)) {
        field.lastFieldState = fieldState;
        notify(fieldSubscribers[name], fieldState, lastFieldState, filterFieldState);
      }
    });
  };

  var isValid = function isValid() {
    return !state.error && Object.keys(state.fields).every(function (key) {
      return !state.fields[key].error && !state.fields[key].submitError;
    });
  };

  var hasSyncErrors = function hasSyncErrors() {
    return state.error || Object.keys(state.fields).some(function (key) {
      return state.fields[key].error;
    });
  };

  var calculateNextFormState = function calculateNextFormState() {
    var fields = state.fields,
        formState = state.formState;

    var fieldKeys = Object.keys(fields);

    // calculate dirty/pristine
    var pristine = fieldKeys.every(function (key) {
      return fields[key].value === fields[key].initial;
    });

    // calculate valid/invalid
    var valid = isValid();

    var validating = state.validating > 0;
    if (pristine === formState.pristine && valid === formState.valid && state.error === formState.error && state.lastFormState && state.lastFormState.validating === validating && state.lastFormState.values === formState.values && state.lastFormState.active === formState.active) {
      return state.lastFormState;
    }

    var active = formState.active,
        initialValues = formState.initialValues,
        submitting = formState.submitting,
        submitError = formState.submitError,
        submitFailed = formState.submitFailed,
        submitSucceeded = formState.submitSucceeded,
        values = formState.values;

    return {
      active: active,
      dirty: !pristine,
      error: state.error,
      initialValues: initialValues,
      invalid: !valid,
      pristine: pristine,
      submitting: submitting,
      submitError: submitError,
      submitFailed: submitFailed,
      submitSucceeded: submitSucceeded,
      valid: valid,
      validating: validating,
      values: values
    };
  };

  var callDebug = function callDebug() {
    return debug && process.env.NODE_ENV !== 'production' && debug(safeFormStateCast(state.formState), Object.keys(state.fields).reduce(function (result, key) {
      result[key] = safeFieldStateCast(state.fields[key]);
      return result;
    }, {}));
  };

  var notifyFormListeners = function notifyFormListeners() {
    callDebug();
    if (inBatch) {
      return;
    }
    var lastFormState = state.lastFormState;

    var nextFormState = calculateNextFormState();
    if (nextFormState !== lastFormState) {
      state.lastFormState = nextFormState;
      notify(state.subscribers, nextFormState, lastFormState, filterFormState);
    }
  };

  // generate initial error (even with no fields yet) if we need to
  runValidation();
  // runValidation(() => {
  //   notifyFieldListeners()
  //   notifyFormListeners()
  // })

  var api = {
    batch: function batch(fn) {
      inBatch = true;
      fn();
      inBatch = false;
      notifyFieldListeners();
      notifyFormListeners();
    },

    blur: function blur(name) {
      var fields = state.fields,
          formState = state.formState;

      var previous = fields[name];
      if (previous && previous.active) {
        // can only blur registered active fields
        delete formState.active;
        fields[name] = _extends({}, previous, {
          active: false,
          touched: true
        });
        if (validateOnBlur) {
          runValidation(function () {
            notifyFieldListeners();
            notifyFormListeners();
          });
        } else {
          notifyFieldListeners();
          notifyFormListeners();
        }
      }
    },

    change: function change(name, value) {
      var fields = state.fields;

      if (fields[name] && fields[name].value !== value) {
        fields[name].value = value;
        state.formState.values = setIn(state.formState.values, name, value) || {};
        notifyFieldListeners();
        notifyFormListeners();
        if (validateOnBlur) {
          notifyFieldListeners();
          notifyFormListeners();
        } else {
          runValidation(function () {
            notifyFieldListeners();
            notifyFormListeners();
          });
        }
      }
    },

    focus: function focus(name) {
      var field = state.fields[name];
      if (field && !field.active) {
        state.formState.active = name;
        field.active = true;
        field.visited = true;
        notifyFieldListeners();
        notifyFormListeners();
      }
    },

    getState: function getState() {
      return safeFormStateCast(state.formState);
    },

    initialize: function initialize(values) {
      var fields = state.fields,
          formState = state.formState;

      formState.initialValues = values;
      formState.values = values;
      Object.keys(fields).forEach(function (key) {
        var field = fields[key];
        var value = getIn(values, key);
        field.value = value;
        field.initial = value;
        field.touched = false;
        field.visited = false;
      });
      runValidation(function () {
        notifyFieldListeners();
        notifyFormListeners();
      });
    },

    registerField: function registerField(name, subscriber) {
      var subscription = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
      var validate = arguments[3];

      if (!state.fieldSubscribers[name]) {
        state.fieldSubscribers[name] = { index: 0, entries: {} };
      }
      var index = state.fieldSubscribers[name].index++;

      // save field subscriber callback
      state.fieldSubscribers[name].entries[index] = {
        subscriber: memoize(subscriber),
        subscription: subscription
      };

      if (!state.fields[name]) {
        // create initial field state
        var initial = state.formState.initialValues ? getIn(state.formState.initialValues, name) : undefined;
        state.fields[name] = {
          active: false,
          blur: function blur() {
            return api.blur(name);
          },
          change: function change(value) {
            return api.change(name, value);
          },
          focus: function focus() {
            return api.focus(name);
          },
          initial: initial,
          lastFieldState: undefined,
          name: name,
          touched: false,
          value: initial,
          validators: {},
          visited: false
        };
      }
      if (validate) {
        state.fields[name].validators[index] = validate;
      }

      var sentFirstNotification = false;
      var firstNotification = function firstNotification() {
        var fieldState = publishFieldState(state.formState, state.fields[name]);
        notifySubscriber(subscriber, subscription, fieldState, undefined, filterFieldState, true);
        state.fields[name].lastFieldState = fieldState;
        sentFirstNotification = true;
      };

      runValidation(function () {
        notifyFormListeners();
        if (!sentFirstNotification) {
          firstNotification();
        }
        notifyFieldListeners();
      });

      return function () {
        delete state.fields[name].validators[index];
        delete state.fieldSubscribers[name].entries[index];
        if (!Object.keys(state.fieldSubscribers[name].entries).length) {
          delete state.fieldSubscribers[name];
          delete state.fields[name];
        }
        runValidation(function () {
          notifyFieldListeners();
          notifyFormListeners();
        });
      };
    },

    reset: function reset() {
      api.initialize(state.formState.initialValues || {});
    },

    submit: function submit() {
      var formState = state.formState,
          fields = state.fields;

      if (hasSyncErrors()) {
        // mark all fields as touched
        Object.keys(fields).forEach(function (key) {
          fields[key].touched = true;
        });
        notifyFieldListeners();
        return; // no submit for you!!
      }
      var resolvePromise = void 0;
      var completeCalled = false;
      var complete = function complete(errors) {
        formState.submitting = false;
        if (errors && Object.keys(errors).length) {
          formState.submitFailed = true;
          formState.submitSucceeded = false;
          Object.keys(fields).forEach(function (key) {
            fields[key].submitError = errors && getIn(errors, key);
          });
          formState.submitError = errors[FORM_ERROR];
        } else {
          Object.keys(fields).forEach(function (key) {
            delete fields[key].submitError;
          });
          delete formState.submitError;
          formState.submitFailed = false;
          formState.submitSucceeded = true;
        }
        notifyFormListeners();
        notifyFieldListeners();
        completeCalled = true;
        if (resolvePromise) {
          resolvePromise();
        }
      };
      formState.submitting = true;
      formState.submitFailed = false;
      formState.submitSucceeded = false;
      if (onSubmit.length === 2) {
        // onSubmit is expecting a callback, first try synchronously
        onSubmit(formState.values, complete);
        if (!completeCalled) {
          // must be async, so we should return a Promise
          notifyFormListeners(); // let everyone know we are submitting
          return new Promise(function (resolve) {
            resolvePromise = resolve;
          });
        }
      } else {
        // onSubmit is either sync or async with a Promise
        var result = onSubmit(formState.values);
        if (result && isPromise(result)) {
          // onSubmit is async with a Promise
          notifyFormListeners(); // let everyone know we are submitting
          return result.then(complete);
        } else {
          // onSubmit is sync
          complete(result);
        }
      }
    },

    subscribe: function subscribe(subscriber, subscription) {
      if (!subscriber) {
        throw new Error('No callback given.');
      }
      if (!subscription) {
        throw new Error('No subscription provided. What values do you want to listen to?');
      }
      var memoized = memoize(subscriber);
      var formState = state.formState,
          subscribers = state.subscribers;

      var index = subscribers.index++;
      subscribers.entries[index] = {
        subscriber: memoized,
        subscription: subscription
      };
      var valid = !state.error;
      var stateWithError = _extends({}, formState, {
        error: state.error,
        invalid: !valid,
        valid: valid
      });
      notifySubscriber(memoized, subscription, stateWithError, stateWithError, filterFormState, true);
      return function () {
        delete subscribers.entries[index];
      };
    }
  };
  return api;
};

//

exports.createForm = createForm;
exports.FORM_ERROR = FORM_ERROR;
exports.version = version;
exports.formSubscriptionItems = formSubscriptionItems;
exports.fieldSubscriptionItems = fieldSubscriptionItems;
